
/*################################################################################################
 *##################                                                            ##################
 *##################            Main program - Compilation/Execution            ##################
 *##################                                                            ##################
 *##################                      Implementation                        ##################
 *##################                                                            ##################
 *################################################################################################
 */


/*
 * cxcl: Interpreter for the XCLE language, based on the libXCLstd
 * builtins implementation
 *
 * Copyright (C) 2000 Yann LANDRIN-SCHWEITZER
 *
 *
 * This library is free software;  you can redistribute it and/or modify it
 * under  the terms of the  GNU Lesser General Public License  as published
 * by the  Free Software Foundation;  either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This  library  is  distributed  in  the hope that it will be useful, but
 * WITHOUT   ANY   WARRANTY;   without   even   the   implied  warranty  of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library;  if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 * Author:   Yann LANDRIN-SCHWEITZER
 * Contact:  varkhan@free.fr
 * Homepage: http://varkhan.free.fr/
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include <DbgLog.h>

#include <XCLE/Splash.h>

#include <XCLE/Containers.h>

#include <XCLE/Obj_generic.h>
#include <XCLE/Obj_void.h>
#include <XCLE/Obj_intg.h>
#include <XCLE/Obj_fltp.h>
#include <XCLE/Obj_strg.h>
#include <XCLE/Obj_list.h>
#include <XCLE/Obj_code.h>

#include <XCLE/Mod_exec.h>
#include <XCLE/Mod_fileIO.h>
#include <XCLE/Mod_parse.h>
#include <XCLE/Mod_write.h>


#include <xclutils.h>



XCLE_ParseCtx pctx ;
XCLE_ParseCtx XCL_parse_ctx = NULL ;
XCLE_ExecCtx  ectx ;
XCLE_ExecCtx  XCL_exec_ctx = NULL ;


//#define BASE_MODULE "libXCLstd.1.so"
//#define DEFAULT_MODULES "libXCLstd_stack.1.so:libXCLstd_flow.1.so:libXCLstd_math.1.so:libXCLstd_list.1.so:libXCLstd_misc.1.so"

#ifndef DEFAULT_MODULES
#define DEFAULT_MODULES "libXCLstd.so"
#endif

//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////   Usage routines
/////



void splash(void) {
	printf("cxcl: eXtensible Concatenative Language interpreter, based on the XCLE library\n") ;
	printf("cxcl and the XCLE library are both (C) 2000-2005 Yann Landrin-Schweitzer, under licence GPL.\n") ;
}

void usage(void) {
	printf("Usage: cxcl -h | -H | -H name | [options] -f file | [options] code-block ...\n") ;
	printf("      -h       : prints the help screen and returns\n") ;
}

void helpscreen() {
	printf("\n") ;
	splash() ;
	printf("\n") ;
	printf("Usage: cxcl -h | [-L] [-l module] | -H | -H name | [options] -f file | [options] code-block ...\n") ;
	printf("      -h | --help       : prints this help and returns\n") ;
	printf("      -v | --version    : prints the version of this program\n") ;
	printf("      -L | --no-defload : do not load default primitives modules\n") ;
	printf("      -l | --load       : loads a binary primitives module\n") ;
	printf("      -H | --prim-list  : prints the list of defined primitives in loaded modules\n") ;
	printf("      -H <name>         : prints information about the primitive <name>\n") ;
	printf("      --prim-info <name>: prints information about the primitive <name>\n") ;
	printf("      --prim-detail     : prints information about all defined primitives\n") ;
	printf("      -f | --file       : introduces a file name to be loaded\n") ;
	printf("\n") ;
	printf("    Options: a leading '-' followed by one or more characters among\n") ;
	printf("      c  *     : prints 'echo' comments in read file (started by ##)\n") ;
	printf("      C  *     : don't print 'echo' comments in read file\n") ;
	printf("      r        : prints error reports for each code-block\n") ;
	printf("      R        : don't print error reports\n") ;
	printf("      s        : prints all stack levels for each code-block\n") ;
	printf("      S        : prints all stack levels at the end of execution\n") ;
	printf("      p        : prints the first stack level for each code-block\n") ;
	printf("      P        : prints the first stack level at the end of execution\n") ;
	printf("      0        : absolutely quiet execution - usefull for stack preload\n") ;
	printf("                   or verbosity reset\n") ;
	printf("    (*: only meaningfull when used with \"-f file\")\n") ;
	printf("\n") ;
	printf("    Syntax:\n") ;
	printf("        Files must contain one code-block per line (use \\ before an \\n to wrap long lines)\n") ;
	printf("        Lines whose first character is # are comments, silently ignored, or echoed when the\n") ;
	printf("        TWO first characters are # and the -c flag is set.\n") ;
	printf("        A code-block is a space-separated list of\n") ;
	printf("      numbers      : +3.141526e00  /  -12345 \n") ;
	printf("      strings      : \"hello world\"\n") ;
	printf("      instructions : <HELP>  /  <*> \n") ;
	printf("      barewords: interpretated in order as numbers, then instructions, and strings\n") ;
	printf("                   if everything else failed\n") ;
	printf("      lists    : [ foo bar 10 2 </> ]\n") ;
	printf("                   ( any of the above, separated by spaces and inside square brackets )\n") ;
	printf("\n") ;
	printf("      To get the list of known built-ins, type:\n") ;
	printf("                 cxcl -H\n") ;
	printf("\n") ;
	printf("      For specific help about some built-in, type:\n") ;
	printf("                 cxcl -H built-in_name\n") ;
	printf("\n") ;
}


// Quick hack to get the infos
struct S_XCLE_ParseCtx {
	unsigned long reg_numb ;
	unsigned char * * reg_name ;
	unsigned long * reg_size ;
	XCLE_CodeDef * * reg_tabl ;
} ;

XCLE_CodeDef * _XCLE_ParseCtx_CodeFind(XCLE_ParseCtx contx, char * str) ;

void bltnlist() {
	unsigned long i = 0 ;
	unsigned long j ;
	while(i<pctx->reg_numb) {
		printf("Registry \"%s\"\n",pctx->reg_name[i]) ;
		j = 0 ;
		while(j<pctx->reg_size[i]) {
			printf("\t%s\t\t%s\n",pctx->reg_tabl[i][j].name,pctx->reg_tabl[i][j].desc) ;
			j ++ ;
		}
		printf("\n") ;
		i ++ ;
	}
}

void bltnabout(char * name) {
	XCLE_CodeDef * blt ;
	unsigned long i ;
	unsigned long j ;
	char * str ;
	blt = _XCLE_ParseCtx_CodeFind(pctx,name) ;
	printf("<%s>\n",name) ;
	if(blt==NULL) {
		printf("\tUNKNOWN\n") ;
	}
	else {
		printf("\t%s\n",blt->desc) ;
		printf("\n") ;
		i = blt->argc ;
		printf("\tArguments: %lu\n",i) ;
		while(i>0) {
			i -- ;
			printf("\t\t_ ") ;
			j = 0 ;
			while(j<5) {
				if(blt->argt[i]&(1<<j)) {
					str = XCLE_ObjectTypeName(1<<j) ;
					if((str==NULL)||(str[0]==0)) break ;
					printf("%s",str) ;
					break ;
				}
				j ++ ;
			}
			j ++ ;
			while(j<5) {
				if(blt->argt[i]&(1<<j)) {
					str = XCLE_ObjectTypeName(1<<j) ;
					if((str==NULL)||(str[0]==0)) break ;
					printf(" | %s",str) ;
				}
				j ++ ;
			}
			printf("\n") ;
		}
		printf("\n") ;
		i = blt->retc ;
		if(i!=255) {
			printf("\tReturns:   %lu\n",i) ;
			while(i>0) {
				i -- ;
				printf("\t\t_ ") ;
				j = 0 ;
				while(j<5) {
					if(blt->rett[i]&(1<<j)) {
						str = XCLE_ObjectTypeName(1<<j) ;
						if((str==NULL)||(str[0]==0)) break ;
						printf("%s",str) ;
						break ;
					}
					j ++ ;
				}
				j ++ ;
				while(j<5) {
					if(blt->rett[i]&(1<<j)) {
						str = XCLE_ObjectTypeName(1<<j) ;
						if((str==NULL)||(str[0]==0)) break ;
						printf(" | %s",str) ;
					}
					j ++ ;
				}
				printf("\n") ;
			}
			printf("\n") ;
		}
		else {
			printf("\tReturns:   not specified\n") ;
			printf("\n") ;
		}
	}
}


void bltndetail() {
	unsigned long i = 0 ;
	unsigned long j ;
	unsigned long k ;
	unsigned long l ;
	char * str ;
	XCLE_CodeDef blt ;
	while(i<pctx->reg_numb) {
		printf("Registry \"%s\"\n",pctx->reg_name[i]) ;
		printf("\n") ;
		j = 0 ;
		while(j<pctx->reg_size[i]) {
			blt = pctx->reg_tabl[i][j] ;
			printf("<%s>\n",blt.name) ;
			printf("\t%s\n",blt.desc) ;
			printf("\n") ;
			k = blt.argc ;
			printf("\tArguments: %lu\n",i) ;
			while(k>0) {
				k -- ;
				printf("\t\t_ ") ;
				l = 0 ;
				while(l<5) {
					if(blt.argt[k]&(1<<l)) {
						str = XCLE_ObjectTypeName(1<<l) ;
						if((str==NULL)||(str[0]==0)) break ;
						printf("%s",str) ;
						break ;
					}
					l ++ ;
				}
				l ++ ;
				while(l<5) {
					if(blt.argt[k]&(1<<l)) {
						str = XCLE_ObjectTypeName(1<<l) ;
						if((str==NULL)||(str[0]==0)) break ;
						printf(" | %s",str) ;
					}
					l ++ ;
				}
				printf("\n") ;
			}
			printf("\n") ;
			k = blt.retc ;
			if(k!=255) {
				printf("\tReturns:   %lu\n",i) ;
				while(k>0) {
					k -- ;
					printf("\t\t_ ") ;
					l = 0 ;
					while(l<5) {
						if(blt.rett[k]&(1<<l)) {
							str = XCLE_ObjectTypeName(1<<l) ;
							if((str==NULL)||(str[0]==0)) break ;
							printf("%s",str) ;
							break ;
						}
						l ++ ;
					}
					l ++ ;
					while(l<5) {
						if(blt.rett[k]&(1<<l)) {
							str = XCLE_ObjectTypeName(1<<l) ;
							if((str==NULL)||(str[0]==0)) break ;
							printf(" | %s",str) ;
						}
						l ++ ;
					}
					printf("\n") ;
				}
				printf("\n") ;
			}
			else {
				printf("\tReturns:   not specified\n") ;
				printf("\n") ;
			}
			printf("\n") ;
			j ++ ;
		}
		printf("\n") ;
		i ++ ;
	}
}

void version() {
	printf("cxcl %s (C) 2000 Yann Landrin-Schweitzer\n",VERSION) ;
	printf("Uses the XCLE library, %s\n",XCLE_copyright()) ;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
/////
/////   MAIN
/////

int main(int argc, char * * argv) {
	int i = 1 ;
	int j = 0 ;
	int verb = REPORT_ALL|STACK_ALL ;
	int defload = 1 ;
	int modnum = 0 ;
	char * modules[256] ;
	XCLE_Stack stk ;
	XCLE_Hash hsh ;

	if(argc<2) {
		usage() ;
		return 0 ;
	}
	
	while(i<argc) {

		if( (!strncmp(argv[i],"-h",3)) || (!strncmp(argv[i],"--help",7)) ) {
			helpscreen() ;
			return 0 ;
		}

		else if( (!strncmp(argv[i],"-v",3)) || (!strncmp(argv[i],"--version",10)) ) {
			version() ;
			return 0 ;
		}

		else if( (!strncmp(argv[i],"-u",2)) ) {
			j = 2 ;
			DbgLog_UpDbgLvl("",1) ;
			DbgLog_UpDbgLvl(NULL,1) ;
			DbgLog("main","MAIN",0,"Error Tracking active") ;
			while(argv[i][j]!='\0') {
				if((argv[i][j]>'0')&&(argv[i][j]<='9')) {
					DbgLog_UpDbgLvl("",1<<(argv[i][j]-'0')) ;
					DbgLog_UpDbgLvl(NULL,1<<(argv[i][j]-'0')) ;
					DbgLog("main","Main",argv[i][j]-'0',"Debug level enabled") ;
				}
				else if((argv[i][j]>='A')&&(argv[i][j]<='Z')) {
					DbgLog_UpDbgLvl("",1<<(argv[i][j]-'A'+10)) ;
					DbgLog_UpDbgLvl(NULL,1<<(argv[i][j]-'A'+10)) ;
					DbgLog("main","Main",argv[i][j]-'A'+10,"Debug level enabled") ;
				}
				j ++ ;
			}
		}

		else if( (!strncmp(argv[i],"-L",3)) || (!strncmp(argv[i],"--no-defload",13)) ) {
			defload = 0 ;
		}

		else if( (!strncmp(argv[i],"-l",3)) || (!strncmp(argv[i],"--load",7)) ) {
			i ++ ;
			if(i<argc) {
				if(modnum>255) {
					fprintf(stderr,"\"-l\" option cannot currently handle more than 256 modules\n") ;
					usage() ;
					return 1 ;
				}
				modules[modnum] = argv[i] ;
				modnum ++ ;
			}
			else {
				fprintf(stderr,"\"-l\" option used without module name\n") ;
				usage() ;
				return 1 ;
			}
		}

		// Not a switch
		else {
			break ;
		}
		
		i ++ ;
	}

	pctx = XCLE_ParseCtxAlloc() ;
	if(pctx==NULL) {
		DbgLog("main","MAIN",1,"Could not obtain parsing context") ;
		return 1 ;
	}
	XCL_parse_ctx = pctx ;

	// Loading default module and additionnal
	if(!defload && !modnum) {
		fprintf(stderr,"No instruction module loaded\n") ;
		usage() ;
		return 1 ;
	}
	if(defload) {
		char * defmodules ;
		/*
		if(XCLE_CodeLoad(pctx,BASE_MODULE)==-1) {
			fprintf(stderr,"Loading default instruction set \"%s\" failed",BASE_MODULE) ;
			return 1 ;
		}
		*/
		defmodules = getenv("CXCL_DEFAULT_MODULES") ;
		if(defmodules==NULL) {
			defmodules = DEFAULT_MODULES ;
		}
		j = 0 ;
		while(defmodules[j]!='\0') {
			unsigned long k = j ;
			char * modname ;
			while((defmodules[k]!='\0')&&(defmodules[k]!=':')) k ++ ;
			modname = malloc((k+1)*sizeof(char)) ;
			if(modname==NULL) {
				fprintf(stderr,"Could not allocate memory for module name\n") ;
				return 2 ;
			}
			k = 0 ;
			while((defmodules[j]!='\0')&&(defmodules[j]!=':')) {
				modname[k] = defmodules[j] ;
				j ++ ;
				k ++ ;
			}
			modname[k] = '\0' ;
			if(k>0) if(XCLE_CodeLoad(pctx,modname)==-1) {
				fprintf(stderr,"Loading default instruction set \"%s\" failed\n",modname) ;
				return 1 ;
			}
			free(modname) ;
		}
	}
	j = 0 ;
	while(j<modnum) {
		if(XCLE_CodeLoad(pctx,modules[j])==-1) {
			fprintf(stderr,"Loading module \"%s\" failed\n",modules[j]) ;
			return 1 ;
		}
		j ++ ;
	}

	// Parsing -H switch: loaded instruction info
	if( (i<argc) && ( (!strncmp(argv[i],"-H",3)) ) ) {
		i ++ ;
		if(argc>i) {
			while(i<argc) {
				bltnabout(argv[i]) ;
				printf("\n") ;
				i ++ ;
			}
		}
		else {
			bltnlist() ;
		}
		return 0 ;
	}
	else if( (i<argc) && (!strncmp(argv[i],"--prim-list",10)) ) {
		bltnlist() ;
		return 0 ;
	}
	else if( (i<argc) && (!strncmp(argv[i],"--prim-info",10)) ) {
		i ++ ;
		if(argc>i) {
			while(i<argc) {
				bltnabout(argv[i]) ;
				printf("\n") ;
				i ++ ;
			}
			return 0 ;
		}
	}
	else if(  (i<argc) && (!strncmp(argv[i],"--prim-detail",14)) ) {
		bltndetail() ;
		return 0 ;	
	}
	if(!(i<argc)) {
		fprintf(stderr,"Too few arguments\n") ;
		usage() ;
		return 1 ;
	}


	stk = XCLE_StackAlloc(10) ;
	hsh = XCLE_HashAlloc(10) ;
	ectx = XCLE_ExecCtxAlloc(10) ;
	if(ectx==NULL) {
		DbgLog("main","MAIN",1,"Could not obtain execution context") ;
		return 1 ;
	}
	XCL_exec_ctx = ectx ;


	while(i<argc) {
		int nverb = 0 ;
		int fload = 0 ;
		while( (i<argc) && (argv[i][0] == '-') && (argv[i][1]!=0) ) {
			if(!strncmp(argv[i],"--file",9)) {
				fload = 1 ;
				i ++ ;
				continue ;
			}
			nverb = verb ;
			j = 1 ;
			while(argv[i][j]!=0) {
				switch(argv[i][j]) {
					case 'c':
						nverb |= PRINT_COM ;
						break ;
					case 'C':
						nverb &= ~PRINT_COM ;
						break ;
					case 'p':
						nverb &= ~PRINT_END ;
						nverb |= PRINT_ALL ;
						break ;
					case 'P':
						nverb &= ~PRINT_ALL ;
						nverb |= PRINT_END ;
						break ;
					case 'r':
						nverb |= REPORT_ALL ;
						break ;
					case 'R':
						nverb &= ~REPORT_ALL ;
						break ;
					case 's':
						nverb &= ~STACK_END ;
						nverb |= STACK_ALL ;
						break ;
					case 'S':
						nverb &= ~STACK_ALL ;
						nverb |= STACK_END ;
						break ;
					case '0':
						nverb = 0 ;
						break ;
					case 'f':
						fload = 1 ;
						break ;
					default:
						j = -1 ;
				}
				if(j<0) break ;
				j ++ ;
			}
			if(j<=1) break ;
			if(j>1) { // we never encountered an invalid char
				verb = nverb ;
				i ++ ;
			}
		} // either the arg did not begin with - or contained an invalid char
		if(fload) {
			if(i<argc) {
				FileExec(argv[i],verb,stk,hsh,pctx,ectx) ;
			}
			else {
				fprintf(stderr,"\"-f\" option used without filename\n") ;
				usage() ;
				return 1 ;
			}
		}
		else {
			if(i<argc) {
				StrExec(argv[i],verb,stk,hsh,pctx,ectx) ;
			}
		}
		i ++ ;
	}
	if(verb&STACK_END) {
		printf("\n") ;
		XCLE_StackWrite(stk,outprint,stderr,outformat,STK_HDR,STK_FMT,STK_TLR,8) ;
		printf("\n") ;
	}
	if(verb&PRINT_END) {
		if(XCLE_StackDepth(stk)>0) XCLE_ObjectWrite(XCLE_StackGet(stk,1),outprint,stdout,outformat) ;
		printf("\n") ;
	}

// Overhead to be sure of clean output with dmalloc
	XCLE_StackFree(stk) ;
	XCLE_HashFree(hsh) ;
	XCLE_ExecCtxFree(ectx) ;

	
	return 0 ;
}



